/*
 * Rahmenwerk-Plug-in "Anlagenstatus"
 * Copyright (C) 2010 BitCtrl Systems GmbH
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.buv.plugin.anlagenstatus.figures;

import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.LabeledContainer;
import org.eclipse.swt.widgets.Display;

import de.bsvrz.buv.plugin.anlagenstatus.parts.AnlagenStatusEditPart;
import de.bsvrz.buv.rw.rw.dav.DaVVerbindung;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Figure zur Anzeige von Online-Daten.
 * 
 * @author BitCtrl Systems GmbH, Steffen Gieseler
 * @version $Id: OnlineDatenFigure.java 32959 2012-02-29 15:05:11Z gieseler $
 */
public abstract class OnlineDatenFigure extends LabeledContainer implements
		ClientReceiverInterface {

	/**
	 * Property f&uuml;r empfangene Datenaktualisierung.
	 */
	public static final String DATEN_UPDATE_PROPERTY = "davUpdate";

	/**
	 * Flag, ob Online-Daten angemeldet sind.
	 */
	private boolean angemeldet;

	/**
	 * Debug-Logger.
	 */
	private final Debug debug = Debug.getLogger();

	/**
	 * Der zugeh&ouml;hroge Edit-Part.
	 */
	private final AnlagenStatusEditPart part;

	/**
	 * Die Datenbeschreigung der Online-Daten.
	 */
	private DataDescription datadescription;

	/**
	 * Gibt die Datenbeschreibung f&uuml;r die Online-Daten zur&uuml;ck.
	 * 
	 * @return Datenbeschreibung
	 */
	public DataDescription getDatadescription() {
		return datadescription;
	}

	/**
	 * Setzt die Datenbeschreibung f&uuml;r die Online-Daten.
	 * 
	 * @param datadescription
	 *            Datenbeschreibung
	 */
	public void setDatadescription(final DataDescription datadescription) {
		this.datadescription = datadescription;
		datenSatzListenerAnmelden();
	}

	/**
	 * Konstruktor.
	 * 
	 * @param parentpart
	 *            Parent-EditPart
	 */
	public OnlineDatenFigure(final AnlagenStatusEditPart parentpart) {
		super();
		part = parentpart;
		// setLayoutManager(new GridLayout(2, false));
		// setFont(parentpart.getFontRegistry().defaultFont());
	}

	/** {@inheritDoc} */
	@Override
	public void removeNotify() {
		if (angemeldet) {
			datenSatzListenerAbmelden();
			angemeldet = false;
		}
		super.removeNotify();
	}

	/**
	 * Meldet die Datensatzlistener an.
	 */
	public void datenSatzListenerAnmelden() {
		final SystemObject[] objects = new SystemObject[1];
		objects[0] = part.getModel().getSystemObject();

		if (objects.length > 0) {
			// Testen, ob Aspekt nicht als Empfnger sondern als Senke
			// angemeldet werden muss.
			// ist der Fall bei "konfigurierenden Parametern" der TLS,
			// Aspekt = "asp.tlsVorgabe"
			// Aspekt = "asp.tlsAbruf"
			boolean receiveAsDrain;
			if ("asp.tlsVorgabe".equals(datadescription.getAspect().toString())
					|| "asp.tlsAbruf".equals(datadescription.getAspect()
							.toString())) {
				receiveAsDrain = true;
			} else {
				receiveAsDrain = false;
			}
			debug.finer("Anmeldung zum Empfang von Daten beim DaV als "
					+ (receiveAsDrain ? "SENKE" : "Empfnger") + " :"
					+ "(Objekt: " + objects[0] + ") (ATG: "
					+ datadescription.getAttributeGroup().toString()
					+ ") (ASP: " + datadescription.getAspect().toString() + ")"
					+ ")");

			// ... und zum empfangen anmelden.
			DaVVerbindung.getInstanz().getDaVVerbindung().subscribeReceiver(
					this, objects, datadescription, new ReceiveOptions(false), // keine
					// Deltas,
					// keine
					// nachgelieferten
					// Daten
					new ReceiverRole(receiveAsDrain)); // empfangen als Senke
			// oder Empfnger
		}

		angemeldet = true;
	}

	/**
	 * Meldet die Datensatzlistener ab.
	 */
	public void datenSatzListenerAbmelden() {
		final SystemObject[] objects = new SystemObject[1];
		objects[0] = part.getModel().getSystemObject();

		DaVVerbindung.getInstanz().getDaVVerbindung().unsubscribeReceiver(this,
				objects, datadescription);
	}

	/**
	 * Aktualisiert die angezeigten Informationen mit den aktuellen Daten der
	 * verwendeten Datenstze.
	 * 
	 * @param results
	 *            ResultDate
	 */
	protected void aktualisiereDaten(final ResultData[] results) {
		for (ResultData result : results) {
			if (result != null && result.hasData()) {
				showDaten(result);
			} else {
				showStatus(result.getDataState().toString());
			}
		}
		firePropertyChange(DATEN_UPDATE_PROPERTY, null, null);
	}

	/** {@inheritDoc} */
	@Override
	public void update(final ResultData[] results) {
		Display.getDefault().asyncExec(new Runnable() {

			public void run() {
				aktualisiereDaten(results);
			}
		});
	}

	/**
	 * Zeigt die Daten an.
	 * 
	 * @param result
	 *            Result
	 */
	public void showDaten(final ResultData result) {
	}

	/**
	 * Zeigt den Daten-Status an.
	 * 
	 * @param status
	 *            Status
	 * 
	 */
	public void showStatus(final String status) {
		removeAll();
		add(new Label(status));
		add(new Label());
	}

}
