/*
 * Rahmenwerk-Plug-in "Anlagenstatus"
 * Copyright (C) 2010 BitCtrl Systems GmbH
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.buv.plugin.anlagenstatus.modell;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;

import de.bsvrz.buv.plugin.anlagenstatus.impl.AnlagenstatusDoModelImpl;
import de.bsvrz.buv.plugin.anlagenstatus.internal.AnlagenStatusKnotenStatus;
import de.bsvrz.buv.plugin.anlagenstatus.parts.AnlagenStatusEditPart;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.tmkextlsglobal.attribute.AttTlsSysKnotenNummer;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Knoten des Anlagenstatus-Modells.
 * 
 * @author BitCtrl Systems GmbH, Steffen Gieseler
 * @version $Id: AnlagenStatusKnoten.java 43864 2013-05-14 11:10:42Z hoesel $
 */
public abstract class AnlagenStatusKnoten extends AnlagenstatusDoModelImpl
		implements PropertyChangeListener {

	/** Anzeige-Label. */
	private final String label;

	/** Die Lokation. */
	private Point location = new Point(0, 0);

	/** Die Gr&ouml;sse der Figur. */
	private Dimension figureSize = new Dimension(0, 0);

	/** Level in der Hierarchie. */
	private final int hLevel;

	/** Listener. */
	private final PropertyChangeSupport listeners = new PropertyChangeSupport(
			this);

	/** Liste der source connections. **/
	private final List<AnlagenStatusConnection> sourceConnections = new ArrayList<AnlagenStatusConnection>();

	/** Liste der target connections. **/
	private final List<AnlagenStatusConnection> targetConnections = new ArrayList<AnlagenStatusConnection>();

	/** Logger. */
	private final Debug logger = Debug.getLogger();

	/** Knotennummer des Ger&auml;tes. */
	private final AttTlsSysKnotenNummer geraetKnotennummer;

	/** Das Anlagenstatusmodell. */
	private final AnlagenStatusModell anlagenStatusModell;

	/** Dei Kinder des Knotens. */
	private final List<AnlagenStatusKnoten> kinder = new ArrayList<AnlagenStatusKnoten>();

	/** Flag, ob sichtbar. */
	private boolean sichtbar = true;

	/** Parent-Knoten. */
	private final AnlagenStatusKnoten parentKnoten;

	/** Der aktuelle Status des Knotens. */
	private AnlagenStatusKnotenStatus knotenStatus = AnlagenStatusKnotenStatus.UNBEKANNT;

	/**
	 * Flag, ob der Knoten angemeldet ist.
	 */
	private boolean online;

	/**
	 * Konstruktor.
	 * 
	 * @param modell
	 *            AnlagenStatusModell
	 * @param parent
	 *            Parent
	 * @param systemObjekt
	 *            SystemObjekt
	 * @param level
	 *            Hierarchielevel
	 */
	public AnlagenStatusKnoten(final AnlagenStatusModell modell,
			final AnlagenStatusKnoten parent, final SystemObjekt systemObjekt,
			final int level) {
		logger.fine("Anlagenstatus-Plugin: Knoten '" + systemObjekt.getName()
				+ "' wird erzeugt");

		this.parentKnoten = parent;
		setSystemObject(systemObjekt.getSystemObject());

		label = systemObjekt.getName();
		hLevel = level;
		anlagenStatusModell = modell;

		geraetKnotennummer = setKnotenNummer();

		erzeugeKinder();
	}

	/**
	 * Gibt das Anlagenstatus-Modell zur&uuml;ck.
	 * 
	 * @return AnlagenStatusModell
	 */
	protected AnlagenStatusModell getAnlagenStatusModell() {
		return anlagenStatusModell;
	}

	/**
	 * F&uuml;gt ein Kind-Element hinzu.
	 * 
	 * @param kind
	 *            AnlagenStatusKnoten
	 */
	public void addKind(final AnlagenStatusKnoten kind) {
		kinder.add(kind);
	}

	/**
	 * Test, ob der Knoten angemeldet ist.
	 * 
	 * @return true, wenn angemeldet, sonst false
	 */
	public boolean isOnline() {
		return online;
	}

	/**
	 * Gibt das Hierarchie-Level zur&uuml;ck.
	 * 
	 * @return Hierarchie-Level
	 */
	public int getlevel() {
		return hLevel;
	}

	/**
	 * Setzt den Knotenstatus.
	 * 
	 * @param knotenStatus
	 *            AnlagenStatusKnotenStatus
	 */
	protected void setKnotenStatus(final AnlagenStatusKnotenStatus knotenStatus) {
		this.knotenStatus = knotenStatus;

	}

	/**
	 * Gibt den aktuellen Knotenstatus zur&uuml;ck.
	 * 
	 * @return aktueller Knotenstatus
	 */
	public AnlagenStatusKnotenStatus getKnotenStatus() {
		return knotenStatus;
	}

	/**
	 * Gibt den Parent-Knoten zur&uuml;k.
	 * 
	 * @return Parent-Knoten
	 */
	public AnlagenStatusKnoten getParentKnoten() {
		return parentKnoten;
	}

	/**
	 * Gibt die Kind-Knoten zur&uuml;ck.
	 * 
	 * @return Liste der Kind-Knoten
	 */
	public List<AnlagenStatusKnoten> getKinder() {
		final List<AnlagenStatusKnoten> klist = new ArrayList<AnlagenStatusKnoten>();

		for (AnlagenStatusKnoten k : kinder) {
			klist.add(k);
		}

		return klist;
	}

	/**
	 * Setzt die Knotennnummer.
	 * 
	 * @return Knotennnummer
	 */
	abstract AttTlsSysKnotenNummer setKnotenNummer();

	/**
	 * Erzeugt die Kindelemente.
	 */
	public void erzeugeKinder() {
	}

	/**
	 * Gibt die Knotennnummer zur&uuml;ck.
	 * 
	 * @return Knotennnummer
	 */
	public String getKnotenNummer() {
		return geraetKnotennummer.toString();
	}

	/**
	 * Gibt das Anzeigelabel zur&uuml;ck.
	 * 
	 * @return Anzeigelabel
	 */
	public String getLabel() {
		return label;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return "AnlagenStatusKnoten(" + label + ")";
	}

	/** {@inheritDoc} */
	@Override
	public void setLocation(final Point iLocation) {
		location = iLocation;
		logger.fine(this + ": setLocation: " + location);
	}

	/**
	 * Berechnet die ben&ouml;tigte H&ouml;he zur Darstellung der Kinder.
	 * 
	 * @return H&ouml;he zur Darstellung der Kinder
	 */
	protected int berechneKinderHoehe() {
		final List<AnlagenStatusKnoten> kliste = getKinder();
		if (kliste.size() > 0) {
			final AnlagenStatusKnoten kindoben = findeKindOben();
			final AnlagenStatusKnoten kindunten = findeKindUnten();
			if (kindoben != null && kindunten != null) {

				return kindunten.getLocation().y + kindunten.getSize().height
						- kindoben.getLocation().y;
			}
		}
		return 0;
	}

	/**
	 * Bestimmt den Kind-Knoten ganz oben.
	 * 
	 * @return Kind-Knoten ganz oben
	 */
	private AnlagenStatusKnoten findeKindOben() {
		int y = 0;
		AnlagenStatusKnoten found = null;
		for (AnlagenStatusKnoten k : getKinder()) {
			if (k.isSichtbar()) {
				if (found == null || k.getLocation().y < y) {
					found = k;
					y = k.getLocation().y;
				}
			}
		}

		return found;
	}

	/**
	 * Bestimmt den Kind-Knoten ganz unten.
	 * 
	 * @return Kind-Knoten ganz unten
	 */
	private AnlagenStatusKnoten findeKindUnten() {
		int y = 0;
		AnlagenStatusKnoten found = null;
		for (AnlagenStatusKnoten k : getKinder()) {
			if (k.isSichtbar()) {
				if (found == null || k.getLocation().y > y) {
					found = k;
					y = k.getLocation().y;
				}
			}
		}

		return found;
	}

	/**
	 * Berechnet die ben&ouml;tigte Breite zur Darstellung der Kinder.
	 * 
	 * @return Breite zur Darstellung der Kinder
	 */
	protected int berechneKinderBreite() {
		final List<AnlagenStatusKnoten> kliste = getKinder();
		if (kliste.size() > 0) {
			final AnlagenStatusKnoten kindlinks = findeKindLinks();
			final AnlagenStatusKnoten kindrechts = findeKindRechts();
			if (kindlinks != null && kindrechts != null) {
				return kindrechts.getLocation().x + kindrechts.getSize().width
						- kindlinks.getLocation().x;
			}
		}
		return 0;
	}

	/**
	 * Bestimmt den Kind-Knoten ganz links.
	 * 
	 * @return Kind-Knoten ganz links
	 */
	private AnlagenStatusKnoten findeKindLinks() {
		int x = 0;
		AnlagenStatusKnoten found = null;
		for (AnlagenStatusKnoten k : getKinder()) {
			if (k.isSichtbar()) {
				if (found == null || k.getLocation().x < x) {
					found = k;
					x = k.getLocation().x;
				}
			}
		}

		return found;
	}

	/**
	 * Bestimmt den Kind-Knoten ganz rechts.
	 * 
	 * @return Kind-Knoten ganz rechts
	 */
	private AnlagenStatusKnoten findeKindRechts() {
		int x = 0;
		AnlagenStatusKnoten found = null;
		for (AnlagenStatusKnoten k : getKinder()) {
			if (k.isSichtbar()) {
				if (found == null || k.getLocation().x > x) {
					found = k;
					x = k.getLocation().x;
				}
			}
		}

		return found;
	}

	/** {@inheritDoc} */
	@Override
	public Point getLocation() {
		final Point rloc = new Point(location.x, location.y);

		if (anlagenStatusModell.isTopDown()) {
			final int kb = berechneKinderBreite();
			if (kb > 0) {
				final AnlagenStatusKnoten kl = findeKindLinks();
				if (kb > figureSize.width) {
					rloc.x = kl.getLocation().x + (kb - figureSize.width) / 2;
				} else {
					rloc.x = kl.getLocation().x + (kb / 2)
							- ((figureSize.width) / 2);
				}
				rloc.x = Math.max(rloc.x, location.x);
			}
		} else {
			final int kh = berechneKinderHoehe();
			if (kh > 0) {
				final AnlagenStatusKnoten kl = findeKindOben();
				if (kh > figureSize.height) {
					rloc.y = kl.getLocation().y + (kh - figureSize.height) / 2;
				} else {
					rloc.y = kl.getLocation().y + (kh / 2)
							- ((figureSize.height) / 2);
				}
				rloc.y = Math.max(rloc.y, location.y);
			}
		}

		return rloc;
	}

	/**
	 * Setzt die Gr&ouml;sse.
	 * 
	 * @param size
	 *            Gr&ouml;sse
	 */
	public void setSize(final Dimension size) {
		if (!isSichtbar()) {
			figureSize = new Dimension(0, 0);
		} else {
			figureSize = size;
		}
		logger.fine(this + ": setSize: " + figureSize);
	}

	/**
	 * Gibt die Gr&ouml;sse zur&uuml;ck.
	 * 
	 * @return Gr&ouml;sse
	 */
	public Dimension getSize() {
		if (!isSichtbar()) {
			figureSize = new Dimension(0, 0);
		}
		return figureSize;
	}

	/**
	 * F&uuml;gt einen Listener f&uuml;r Property-Changes hinzu.
	 * 
	 * @param listener
	 *            PropertyChangeListener
	 */
	public void addPropertyChangeListener(final PropertyChangeListener listener) {
		listeners.addPropertyChangeListener(listener);
	}

	/**
	 * Entfernt einen Listener f&uuml;r Property-Changes.
	 * 
	 * @param listener
	 *            PropertyChangeListener
	 */
	public void removePropertyChangeListener(
			final PropertyChangeListener listener) {
		listeners.removePropertyChangeListener(listener);
	}

	/**
	 * Publiziert ein Property-Change.
	 * 
	 * @param prop
	 *            Name der Property
	 * @param old
	 *            alter Wert
	 * @param newValue
	 *            neuer Wert
	 */
	public void firePropertyChange(final String prop, final Object old,
			final Object newValue) {
		logger.fine(this + ":firePropertyChange: " + prop);
		listeners.firePropertyChange(prop, old, newValue);
	}

	/**
	 * Gibt alle Source-Verbindungen des Knotens zur&uuml;ck.
	 * 
	 * @return Liste der Source-Verbindungen
	 */
	public List<AnlagenStatusConnection> getSourceConnections() {
		return sourceConnections;
	}

	/**
	 * Gibt alle aktiven Target-Verbindungen des Knotens zur&uuml;ck.
	 * 
	 * @return Liste der aktiven Target-Verbindungen
	 */
	public List<AnlagenStatusConnection> getAktiveTargetConnections() {
		return targetConnections;
	}

	/**
	 * Gibt alle Target-Verbindungen des Knotens zur&uuml;ck.
	 * 
	 * @return Liste der Target-Verbindungen
	 */
	public List<AnlagenStatusConnection> getTargetConnections() {
		return targetConnections;
	}

	/**
	 * Setzt eine eingehende Verbindung.
	 * 
	 * @param connection
	 *            AnlagenStatusConnection
	 */
	public void connectInput(final AnlagenStatusConnection connection) {
		targetConnections.add(connection);
	}

	/**
	 * Setzt eine ausgehende Verbindung.
	 * 
	 * @param connection
	 *            AnlagenStatusConnection
	 */
	public void connectOutput(final AnlagenStatusConnection connection) {
		sourceConnections.add(connection);
	}

	/**
	 * Gibt den KnotenStatus zur&uuml;ck.
	 * 
	 * @return KnotenStatus
	 */
	public AnlagenStatusKnotenStatus getZustand() {
		return AnlagenStatusKnotenStatus.UNBEKANNT;
	}

	/**
	 * Test, ob der Knoten aktuell sichtbar ist.
	 * 
	 * @return true wenn sichtbar, sonst false
	 */
	public boolean isSichtbar() {
		return sichtbar;
	}

	/**
	 * Setzt die Sichtbarkeit des Knotens.
	 * 
	 * @param newSichtbar
	 *            sichtbar ja/nein
	 * @param recursive
	 *            auf untergeordnete Elemente anwenden ja/nein
	 */
	public void setSichtbar(final boolean newSichtbar, final boolean recursive) {
		this.sichtbar = newSichtbar;
		if (recursive) {
			if (!newSichtbar) {
				untergeordneteEbeneAusblenden();
			} else {
				untergeordneteEbeneEinblenden();
			}
		}
	}

	/**
	 * Blendet den Knoten aus.
	 */
	public void ausblenden() {
		if (getTargetConnections() != null) {
			for (AnlagenStatusConnection c : getTargetConnections()) {
				c.setSichtbar(false);
			}
		}
		setSichtbar(false, true);
	}

	/**
	 * Blendet den Knoten ein.
	 */
	public void einblenden() {
		if (getTargetConnections() != null) {
			for (AnlagenStatusConnection c : getTargetConnections()) {
				c.setSichtbar(true);
			}
		}
		setSichtbar(true, false);
	}

	/**
	 * Blendet die untergeordnete Ebene ein.
	 */
	public void untergeordneteEbeneEinblenden() {
		if (getSourceConnections() != null) {
			for (AnlagenStatusConnection c : getSourceConnections()) {
				// c.getTarget().setSichtbar(false);
				c.setSichtbar(true);
			}
		}

		if (kinder != null) {
			for (AnlagenStatusKnoten k : kinder) {
				k.setSichtbar(true, true);
			}
		}

		getAnlagenStatusModell().updateShowLevel();
	}

	/**
	 * Blendet die untergeordnete Ebene aus.
	 */
	public void untergeordneteEbeneAusblenden() {
		if (getSourceConnections() != null) {
			for (AnlagenStatusConnection c : getSourceConnections()) {
				// c.getTarget().setSichtbar(false);
				c.setSichtbar(false);
			}
		}

		if (kinder != null) {
			for (AnlagenStatusKnoten k : kinder) {
				k.setSichtbar(false, true);
			}
		}

		getAnlagenStatusModell().updateShowLevel();
	}

	/**
	 * Test, ob der Knoten unsichtbare Kinder besitzt.
	 * 
	 * @return true, wenn der Knoten unsichtbare Kinder besitzt, sonst false
	 */
	public boolean hasUnsichtbareKinder() {
		if (kinder != null) {
			for (AnlagenStatusKnoten k : kinder) {
				if (!k.isSichtbar()) {
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Test, ob der Knoten sichtbare Kinder besitzt.
	 * 
	 * @return true, wenn der Knoten sichtbare Kinder besitzt, sonst false
	 */
	public boolean hasSichtbareKinder() {
		if (kinder != null) {
			for (AnlagenStatusKnoten k : kinder) {
				if (k.isSichtbar()) {
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Aktualisiert den Status des Knotens.
	 */
	protected void aktualisiereStatus() {
		int ok = 0;
		int gestoert = 0;
		int teilgestoert = 0;
		int relevant = 0;

		for (AnlagenStatusKnoten knoten : kinder) {
			if (!knoten.isStatusRelevant()) {
				continue;
			}

			relevant++;
			switch (knoten.getKnotenStatus()) {
			case OK:
				ok++;
				break;

			case GESTOERT:
				gestoert++;
				break;

			case TEILGESTOERT:
				teilgestoert++;
				break;

			default:
				break;
			}
		}

		final AnlagenStatusKnotenStatus neu = bildeStatus(relevant, ok,
				teilgestoert, gestoert);

		setKnotenStatus(neu);

		logger.finer(this + "aktualisiereStatus: firePropertyChange");
		firePropertyChange(AnlagenStatusEditPart.STATUS_CHANGED_PROPERTY, null,
				null);
	}

	/**
	 * Bildet den Status des Edit-Parts.
	 * 
	 * <p>
	 * Der Status ist wie folgt definiert:
	 * </p>
	 * <table>
	 * <tr>
	 * <td>Status</td>
	 * <td>Bedingung</td>
	 * </tr>
	 * <tr>
	 * <td>undefiniert</td>
	 * <td>Der Zustand konnte nicht bestimmt werden</td>
	 * </tr>
	 * <tr>
	 * <td>gest&ouml;rt</td>
	 * <td>Alle untergeordneten Teile sind gest&ouml;rt</td>
	 * </tr>
	 * <tr>
	 * <td>teilgest&ouml;rt</td>
	 * <td>Mindestens eines der untergeordneten Teile ist gest&ouml;rt</td>
	 * </tr>
	 * <tr>
	 * <td>ungest&ouml;rt</td>
	 * <td>Alle untergeordneten Teile sind ungest&ouml;rt</td>
	 * </tr>
	 * </table>
	 * 
	 * @param gesamtAnzahl
	 *            Anzahl der untergeordneten Teile
	 * @param okAnzahl
	 *            Anzahl der untergeordneten Teile mit Status OK
	 * @param teilgestoertAnzahl
	 *            Anzahl der untergeordneten Teile mit Status TEILGESTOERT
	 * @param gestoertAnzahl
	 *            Anzahl der untergeordneten Teile mit Status GESTOERT
	 * @return AnlagenStatusKnotenStatus
	 */
	protected AnlagenStatusKnotenStatus bildeStatus(final int gesamtAnzahl,
			final int okAnzahl, final int teilgestoertAnzahl,
			final int gestoertAnzahl) {

		if (okAnzahl == 0 && teilgestoertAnzahl == 0 && gestoertAnzahl == 0) {
			return AnlagenStatusKnotenStatus.UNBEKANNT;
		}

		if (okAnzahl == gesamtAnzahl) {
			return AnlagenStatusKnotenStatus.OK;
		}

		if (gestoertAnzahl == gesamtAnzahl) {
			return AnlagenStatusKnotenStatus.GESTOERT;
		}

		return AnlagenStatusKnotenStatus.TEILGESTOERT;
	}

	/**
	 * Bestimmt, ob der Knoten relevant f&uuml;r die Statusberechnung ist.
	 * 
	 * @return true, wenn relevant, sonst false
	 */
	public boolean isStatusRelevant() {
		for (AnlagenStatusKnoten knoten : kinder) {
			if (knoten.isStatusRelevant()) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Setzt den Knoten online.
	 */
	public void geheOnline() {
		if (!isOnline()) {
			datenanmelden();
			online = true;
		}

		for (AnlagenStatusKnoten knoten : kinder) {
			if (!knoten.isOnline()) {
				knoten.geheOnline();
				// knoten.addPropertyChangeListener(this);
			}
			knoten.addPropertyChangeListener(this);
		}

		// Knotenstatus-Property damit bei Wiederanmeldung der EditPart
		// aktualisiert wird.
		logger.finer(this + "geheOnline: firePropertyChange");
		firePropertyChange(AnlagenStatusEditPart.STATUS_CHANGED_PROPERTY, null,
				null);
	}

	/**
	 * Setzt den Knoten offline.
	 */
	public void geheOffline() {
		if (isOnline()) {
			datenabmelden();
			online = false;
		}

		for (AnlagenStatusKnoten knoten : kinder) {
			// if (knoten.isOnline()) {
			// knoten.geheOffline();
			// }
			knoten.removePropertyChangeListener(this);
		}

		// Knotenstatus-Property damit bei Wiederanmeldung der EditPart
		// aktualisiert wird.
		logger.finer(this + "geheOffline: firePropertyChange");
		firePropertyChange(AnlagenStatusEditPart.STATUS_CHANGED_PROPERTY, null,
				null);
	}

	/**
	 * Meldet die ben&ouml;tigten Daten am DAV an.
	 */
	protected void datenanmelden() {

	}

	/**
	 * Meldet die angemeldeten Daten am DAV ab.
	 */
	protected void datenabmelden() {

	}

	/** {@inheritDoc} */
	@Override
	public void propertyChange(final PropertyChangeEvent evt) {
		final String prop = evt.getPropertyName();
		if (prop.equals(AnlagenStatusEditPart.STATUS_CHANGED_PROPERTY)) {
			aktualisiereStatus();
		}
	}

}
